/*
 *  interface.c - functions used to detect interfaces
 *
 *  Copyright (C) 1999-2000 Hugo Haas & Etienne Bernard
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "defines.h"

#include <stdlib.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/socket.h>
#ifdef __OS_SUNOS__
#include <sys/sockio.h>
#endif
#include <net/if.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <string.h>
#include <pthread.h>
#include <unistd.h>
#ifdef HAVE_STROPTS_H
#include <stropts.h>
#endif

#ifdef __OS_SUNOS__
#	include <unistd.h>
#	include <stropts.h>
#endif

#include "interface.h"
#include "log.h"
#include "thread.h"
#include "pcap.h"

#define isUP(I) ((I)->flags & IFF_UP)

struct interface *interfaces = NULL;

int countInterfaces = 0;

pthread_t *threadsTable = NULL;

void detectInterfaces(void) {
  extern logDevice *mainLog;
  int tryCount = 16, sock, n;
  struct ifconf ifc;
  struct ifreq *ifr, ifri;
  struct sockaddr_in *sin;
  struct interface *i = NULL, *j;
  char *c;

  /* Open a socket */
  sock = socket(AF_INET, SOCK_DGRAM, 0);
  if (sock < 0) goto error;

  /* Try to get information about all the interfaces */
  ifc.ifc_buf = NULL;
  while (1) {
    ifc.ifc_len = sizeof(struct ifreq) * tryCount;
    ifc.ifc_buf = realloc(ifc.ifc_buf, ifc.ifc_len);
    if (ioctl(sock, SIOCGIFCONF, (char *)&ifc) < 0) goto error;
    if (ifc.ifc_len != sizeof(struct ifreq) * tryCount) break;
    tryCount += 10;
  }
    
  /* Parse results */
  ifr = ifc.ifc_req;
  for (n = 0; n < ifc.ifc_len; n += sizeof(struct ifreq)) {
    /* Get more information about the interface */
    strncpy(ifri.ifr_name, ifr->ifr_name, sizeof(ifr->ifr_name));
	/* The interface must be up, and not loopback */
    if (!ioctl(sock, SIOCGIFFLAGS, &ifri) &&
        ifr->ifr_addr.sa_family == AF_INET && (ifri.ifr_flags & IFF_UP)
		&& !(ifri.ifr_flags & IFF_LOOPBACK)) {
      /* Interface does IP and is up. Add it */
#ifdef DEBUG
      mainLog->doLog(mainLog, "Adding interface %s", ifr->ifr_name);
#endif
      /* Deal with IP aliases */
      if ((c = strchr(ifr->ifr_name, ':')) != NULL)
        *c = '\0';

      /* Find interface entry */
      i = NULL;
      for (j = interfaces; j != NULL; j = j->next)
        if (!strcmp(j->name, ifr->ifr_name)) {
          i = j;
          break;
        }

      /* Get IP address */
      sin = (struct sockaddr_in *)&ifr->ifr_addr;

      if (i) {
        /* Add IP */
        i->ip = (struct in_addr *)realloc(i->ip, sizeof(struct in_addr)*(i->countIP+1));
        i->ip[i->countIP++] = sin->sin_addr;
      } else {
        i = (struct interface *)malloc(sizeof(struct interface));
        /* Get name */
        i->name = strdup(ifr->ifr_name);
        i->ip = (struct in_addr *)malloc(sizeof(struct in_addr));
        i->ip[0] = sin->sin_addr;
        i->countIP = 1;
        i->next = interfaces;
        interfaces = i;
        countInterfaces++;
      }
    }
    ifr++;
  }

  free(ifc.ifc_buf);
  return;

 error:
  mainLog->doLog(mainLog, "WARNING: Unable to detect local interfaces.");
}

void freeInterfaces(void) {
  struct interface *i;

  while (interfaces) {
    i = interfaces;
    interfaces = interfaces->next;
    free(i->name);
    free(i);
  }

  countInterfaces = 0;
}

/* Stop all interface threads */
void stopInterfaceThreads(void) {
  int i;

  /* Cancel the threads */
  for (i = 0; i < countInterfaces; i++) {
    pthread_cancel(threadsTable[i]);
  }

  /* Free threads table */
  free(threadsTable);
}

/* Start all interface threads */
void startInterfaceThreads(struct passwd *user) {
  int j, c = 0;
  struct interface *i;
  struct interfaceThreadParams *params;

  /* Free interface list */
  freeInterfaces();

  /* Detect interfaces */
  detectInterfaces();

  /* Allocate thread table */
  threadsTable = (pthread_t *)malloc(countInterfaces * sizeof(pthread_t));

  /* For each interface, start a thread */
  for (i = interfaces; i != NULL; i = i->next) {
    params = (struct interfaceThreadParams *)malloc(sizeof(struct interfaceThreadParams));
    params->user = user;
    params->name = i->name;
    params->countIP = i->countIP;
    params->ip = (struct in_addr *)malloc(sizeof(struct in_addr)*i->countIP);
    for (j = 0; j < i->countIP; j++) params->ip[j] = i->ip[j];
    runThread(&threadsTable[c], (void *(*)(void *))interfaceThread, (void *)params);
    c++;
  }
}
