/*
 *  main.c - Main file
 *
 *  Copyright (C) 1999-2000 Hugo Haas & Etienne Bernard
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "defines.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pwd.h>
#include <sys/ioctl.h>
#include <sys/termios.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <pthread.h>
#include <unistd.h>
#ifdef HAVE_STROPTS_H
#include <stropts.h>
#endif
#include <netinet/in.h>
#ifndef HAVE_GETOPT_LONG
#	include "getopt.h"
#else
#	include <getopt.h>
#endif

#include "log.h"
#include "configuration.h"
#include "filter.h"
#include "resolv.h"
#include "interface.h"
#include "logging.h"
#include "resolv.h"
#include "pidfile.h"

/* Defaults */
char *configFile = NULL;
int goBackground = 1;

logDevice *mainLog;
logDevice *icmpLog = NULL;
logDevice *tcpLog = NULL;
logDevice *udpLog = NULL;

/* Filter */
struct filterEntry *filter = NULL;

/* Give help about the program's syntax */
void usage() {
  printf("ippl version " VERSION "\n\n");
  printf("\t-n, --nodaemon:\tdo not run as a daemon and write information to stdout\n");
  printf("\t-c, --config <file>:\tspecify an alternate configuration file\n");
  printf("\t-h, --help:\tdisplay this help and exit\n");
  printf("\nPlease report bugs to ippl@via.ecp.fr\n");
}


/* Parse the arguments given on the command line */
void parseArguments(int argc, char *argv[]) {
  int c, optionIndex = 0;
  static struct option longOptions[] =
  {
    { "help", 0, 0, 0 },
    { "nodaemon", 0, &goBackground, 0 },
    { "config", 1, 0, 0 },
    { 0, 0, 0, 0 }
  };

  while ((c = getopt_long(argc, argv, "hnc:", longOptions,
                          &optionIndex)) != -1) {
    switch (c) {
    case 0:
      switch (optionIndex) {
      case 0: usage(); exit(0);
      case 2: configFile = optarg;
      }
      break;
    case 'h': usage(); exit(0);
    case 'n': goBackground = 0; break;
    case 'c': configFile = optarg; break;
    case '?': usage(); exit(1);
    }
  }
}

/* Start all threads */
void startAllThreads() {
  extern unsigned short logProtocols;
  extern char *safeUser;
  struct passwd *user;

  /* Parse configuration file */
  parseConfigFile(configFile);

  if (!safeUser) safeUser = strdup(DEFAULT_USER);

  /* Get uid for user */
  user = getpwnam(safeUser);
  if (!user) {
    mainLog->doLog(mainLog, "WARNING: I cannot find the \"%s\" account.", safeUser);
    logProtocols = 0;
    return;
  }

  if (!strcmp(safeUser, "root")) {
    mainLog->doLog(mainLog, "WARNING: Using root account to run threads. This is NOT safe!");
  }

  if (logProtocols == 0) {
    mainLog->doLog(mainLog, "WARNING: Nothing to log.");
    return;
  }

  /* Init the DNS cache */
  initDNSCache();

  /* Start resolving thread */
  startResolvThread(user);

  /* Start logging thread */
  startLogThread(user);

  /* Start a logging thread on every interfaces */
  startInterfaceThreads(user);
}

/* Kill all threads and cleanly exit */
void die(int sigNum) {

  /* Stop threads */
  stopInterfaceThreads();
  stopResolvThread();
  stopLogThread();

  /* Close log files */
  if (icmpLog) icmpLog->close(icmpLog);
  if (tcpLog) tcpLog->close(tcpLog);
  if (udpLog) udpLog->close(udpLog);

  if (goBackground) {
    (void) removePid(PID_FILE);
  }

  mainLog->doLog(mainLog, "IP Protocols Logger: stopped (signal %d).", sigNum);
  mainLog->close(mainLog);

  exit(0);
}

/* Reload configuration */
void reload(int sigNum) {

  /* Close log files */
  if (icmpLog) { icmpLog->close(icmpLog); free(icmpLog); icmpLog = NULL; }
  if (tcpLog) { tcpLog->close(tcpLog); free(tcpLog); tcpLog = NULL; }
  if (udpLog) { udpLog->close(udpLog); free(udpLog); udpLog = NULL; }

  /* Kill the threads */
  stopInterfaceThreads();
  stopResolvThread();
  stopLogThread();

  /* Restart them */
  startAllThreads();

  mainLog->doLog(mainLog, "IP Protocols Logger: reloaded configuration.");
  signal(SIGHUP, reload);
}

/* Setup signal handling */
void setupSignals(void) {
  signal(SIGTERM, die);
  if (!goBackground) {
    signal(SIGINT, die);
  }
  signal(SIGHUP, reload);
}

/* Go background */
void background(void) {
  int fd;

  /* Check PID file */
  if (checkPid(PID_FILE)) {
    mainLog->doLog(mainLog, "Already running. Exiting...");
    exit(1);
  }

  if (getppid() != 1) {
    signal(SIGTTOU, SIG_IGN);
    signal(SIGTTIN, SIG_IGN);
    signal(SIGTSTP, SIG_IGN);
	switch (fork()) {
	case -1:
      perror("fork");
      exit(1);
	case 0:
	  break;
	default:
	  exit(0);
    }
    fd = open("/dev/tty", O_RDWR);
    if (fd >= 0) {
      ioctl(fd, TIOCNOTTY, (char *) NULL);
      close(fd);
    }
  }

  close(0); close(1); close(2);
  umask(027);
  setsid();

  /* Write PID file */
  if (!writePid(PID_FILE)) {
    mainLog->doLog(mainLog, "Can't write pid. Exiting...\n");
    exit(1);
  }
}

/* This is the main function */
int main(int argc, char *argv[]) {

  /* Parse parameters */
  parseArguments(argc, argv);
  if (!configFile) configFile = strdup(CONFIGURATION_FILE);

  /* Check that we have the root priviledges */
  setuid(0);
  if (geteuid()) {
    printf("You must be super-user to run this program.\n");
    exit(1);
  }

  /* Setup as daemon */
  if (goBackground) {
	/* Open log device */
	mainLog = newSyslogDevice(LOGLEVEL);
	/* Go background */
	background();
  } else {
	/* Open log device */
	mainLog = newStdOutDevice();
  }

  /* Setup signal handling */
  setupSignals();

  /* Parse configuration file and start all threads */
  startAllThreads();
  mainLog->doLog(mainLog, "IP Protocols Logger: started.");

  /* Wait forever - should be optimized */
  while (1) {
    sleep(86400);
  }

  /* Never reached */
  return 1;
}
