/*
    pidfile.c - interact with pidfiles
    Copyright (c) 2000 Etienne Bernard - Hugo Haas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <fcntl.h>

#include "log.h"

extern logDevice *mainLog;

/* readPid
 *
 * Reads the specified pidfile and returns the read pid.
 * 0 is returned if either there's no pidfile, it's empty
 * or no pid can be read.
 */
int readPid(char *pidfile)
{
  FILE *f;
  int pid;

  f = fopen(pidfile, "r");
  if (!f) return 0;

  fscanf(f, "%d", &pid);
  fclose(f);

  return pid; 
}

/* checkPid
 *
 * Reads the pid using read_pid and looks up the pid in the process
 * table (using /proc) to determine if the process already exists. If
 * so 1 is returned, otherwise 0.
 */
int checkPid (char *pidfile)
{
  int pid;

  pid = readPid(pidfile);
  if (pid == 0)
    return 0;
  if (kill(pid, 0) && errno == ESRCH) return 0;

  return 1;
}

/* writePid
 *
 * Writes the pid to the specified file. If that fails 0 is
 * returned, otherwise the pid.
 */
int writePid (char *pidfile)
{
  FILE *f;
  int fd, pid;
  struct flock lock;

  fd = open(pidfile, O_RDWR|O_CREAT, 0644);
  if (fd == -1) goto errfile;

  f = fdopen(fd, "r+");
  if (!f) goto errfile;

  lock.l_type = F_WRLCK;
  lock.l_whence = SEEK_SET;
  lock.l_start = 0;
  lock.l_len = 0;

  if (fcntl(fd, F_SETLK, &lock) == -1) goto errlock;

  pid = getpid();
  fprintf(f, "%d\n", pid);
  fflush(f);

  close(fd);

  return pid;

errfile:
  mainLog->doLog(mainLog, "Cannot open or create %s", pidfile);
  return 0;

errlock:
  mainLog->doLog(mainLog, "Cannot lock %s", pidfile);
  return 0;
}

/* removePid
 *
 * Remove the the specified file. The result from unlink(2)
 * is returned
 */
int removePid (char *pidfile)
{
  return unlink(pidfile);
}
