/*
 *  resolv.c - handle dns cache
 *
 *  Copyright (C) 1999-2000 Hugo Haas & Etienne Bernard
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <pthread.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <netinet/in.h>
#include <unistd.h>
#include <grp.h>

#include "resolv.h"
#include "thread.h"
#include "filter.h"

/* Threads */
pthread_t resolvThread;

/* Mutexes */
pthread_mutex_t dnsMutex = PTHREAD_MUTEX_INITIALIZER;

/* Data structure */
typedef struct {
  uint32_t ip;
  char *name;
  short dirty;
} hostEntry;

hostEntry hostTable[TABLE_SIZE];

/* Initialise DNS cache */
void initDNSCache(void) {
  int i;

  pthread_mutex_lock(&dnsMutex);

  for (i = 0; i < TABLE_SIZE; i++) {
    hostTable[i].ip = 0;
    if (hostTable[i].name != NULL) {
      free(hostTable[i].name);
      hostTable[i].name = NULL;
    }
  }

  pthread_mutex_unlock(&dnsMutex);
}

/* Perform a DNS lookup */
int performLookup(char *host, uint32_t in) {
  struct in_addr i;
  struct hostent *he;

  i.s_addr = in;
  he = gethostbyaddr((char *)&i, sizeof(struct in_addr), AF_INET);
  if (he) {
    strncpy(host, he->h_name, HOST_LENGTH-24);
    return 1;
  }
  return 0;
}

/* Performs a DNS query with caching
 *
 * Cache uses a hashtable with open addressing and double hashing. It
 * has LEVELS levels. Each entry has a dirty flag to help to decide
 * which entry should be discarded if all the levels are occupied.
 */
int hostLookup(char *host, uint32_t in) {
  uint16_t key, key1, key2, level = 0;
  int result;

  pthread_mutex_lock(&dnsMutex);

  /* First : check if the value is stored in the hash table or do a
     lookup if a slot is empty */
  key = key1 = HASH1(in);
  key2 = HASH2(in);

  do {
    if (hostTable[key].ip == 0) goto lookup;
    if (hostTable[key].ip == in) {
      strcpy(host, hostTable[key].name);
      hostTable[key].dirty = 0;
      goto end;
    }
    key = (key + key2) % TABLE_SIZE;
  } while (++level == LEVELS);

  /* Second : if we are here, it means that all the slots are full and
     none of them contains the search item. Look for an entry to
     replace. Take the first dirty one. */

  key = key1;
  level = 0;
  do {
    if (hostTable[key].dirty)
      goto lookup;
    else
      hostTable[key].dirty = 1;
    key = (key + key2) % TABLE_SIZE;
  } while (++level == LEVELS);

  /* Perform a lookup and store the result in the hash table */
 lookup:
  result = performLookup(host, in);
  hostTable[key].ip = in;
  hostTable[key].dirty = 0;
  if (hostTable[key].name != NULL) free(hostTable[key].name);
  hostTable[key].name = strdup(host);

  pthread_mutex_unlock(&dnsMutex);
  return result;

 end:
  pthread_mutex_unlock(&dnsMutex);
  return 1;
}

/* Main resolving thread loop */
void resolvLoop(struct passwd *user) {
  extern unsigned int dnsExpire;

#ifdef __OS_LINUX__
  /* Switch user */
  if (user) {
    setgid(user->pw_gid);
    initgroups(user->pw_name, user->pw_gid);
    setuid(user->pw_uid);    
  }                          
#endif

  /* Enable cancelation. We must use the DEFERRED mode, so that we can
     not be canceled while holding the dnsMutex */
  pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
  pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);

  while (1) {
    sleep(dnsExpire);
    initDNSCache();
    refreshFilter();
    /* Test for pending cancelation */
    pthread_testcancel();
  }
}

/* Start resolving thread */
void startResolvThread(struct passwd *user) {
  runThread(&resolvThread, (void *(*)(void *))resolvLoop, (void *)user);
}

/* Stop resolving thread */
void stopResolvThread(void) {
  pthread_cancel(resolvThread);
}
